#!/usr/bin/env python

import curses
import logging
import os
import pickle
import sys

import game
import gameworld
import screen
import userinput


ACTION_QUIT = 1
ACTION_EDIT = 2
ACTION_SAVE = 3
ACTION_COLOR_0 = 4
ACTION_COLOR_1 = 5
ACTION_COLOR_2 = 6
ACTION_COLOR_3 = 7
ACTION_COLOR_4 = 8
ACTION_COLOR_5 = 9
ACTION_COLOR_6 = 10
ACTION_COLOR_7 = 11
ACTION_BACKSPACE = 12
ACTION_MENU = 13
ACTION_FRAME_PREV = 14
ACTION_FRAME_NEXT = 15
ACTION_COLOR_PREV = 16
ACTION_COLOR_NEXT = 17
ACTION_DUPLICATE = 18
ACTION_DELETE_FRAME = 19

MODE_MENU = 1
MODE_EDIT = 2
BINDINGS_MENU = {
    ord('q'): ACTION_QUIT,
    ord('s'): ACTION_SAVE,
    ord('d'): ACTION_DUPLICATE,
    ord('l'): ACTION_DELETE_FRAME,
    ord('0'): ACTION_COLOR_0,
    ord('1'): ACTION_COLOR_1,
    ord('2'): ACTION_COLOR_2,
    ord('3'): ACTION_COLOR_3,
    ord('4'): ACTION_COLOR_4,
    ord('5'): ACTION_COLOR_5,
    ord('6'): ACTION_COLOR_6,
    ord('7'): ACTION_COLOR_7,
    11: ACTION_FRAME_PREV,  # ^k
    12: ACTION_FRAME_NEXT,  # ^l
    22: ACTION_COLOR_PREV,  # ^v
    2: ACTION_COLOR_NEXT,  # ^b
    27: userinput.ESC,      # ESC
}
BINDINGS_EDIT = {
    24: ACTION_MENU,        # ^x
    11: ACTION_FRAME_PREV,  # ^k
    12: ACTION_FRAME_NEXT,  # ^l
    22: ACTION_COLOR_PREV,  # ^v
    2: ACTION_COLOR_NEXT,  # ^b
    curses.KEY_UP: userinput.UP,
    curses.KEY_DOWN: userinput.DOWN,
    curses.KEY_LEFT: userinput.LEFT,
    curses.KEY_RIGHT: userinput.RIGHT,
    curses.KEY_BACKSPACE: ACTION_BACKSPACE,
    curses.KEY_DC: ACTION_BACKSPACE,
    127: ACTION_BACKSPACE,
}
TYPEABLE_CHARS = set()
for c in xrange(32, 127):
    BINDINGS_EDIT[c] = c
    TYPEABLE_CHARS.add(c)


class FillRect(screen.Figure):
    def __init__(self, ch, color, *args, **kwargs):
        super(FillRect, self).__init__(*args, **kwargs)
        self.rect_char = ch
        self.rect_attr = curses.color_pair(color)
    def get_char_and_attr(self, x, y):
        return self.rect_char, self.rect_attr


class Border(screen.Figure):
    def __init__(self, *args, **kwargs):
        super(Border, self).__init__(*args, **kwargs)
        self.cur_color = 0

    def get_char_and_attr(self, x, y):
        if x == self.width - 1 and y == self.height - 1:
            return ord('+'), self.cur_color
        if x == self.width - 1:
            return ord('|'), self.cur_color
        if y == self.height - 1:
            return ord('-'), self.cur_color
        return None, None


class Cursor(screen.Figure):
    def __init__(self, *args, **kwargs):
        super(Cursor, self).__init__(1, 1, 0, 0, *args, **kwargs)
        self.cur_color = 0
    def get_char_and_attr(self, x, y):
        return 35, self.cur_color


class MenuBar(screen.Figure):
    def __init__(self, *args, **kwargs):
        super(MenuBar, self).__init__(79, 1, 0, 0, *args, **kwargs)
        self.mode = None
        self.cur_color = 0
        self.dirty = False
        self.frame_id = 0
        self.frame_count = 1
        self.update(None, 0, False)
        self._str = ''

    def update(self, mode=None, cur_color=None, dirty=None, frame_id=None, frame_count=None):
        if mode is not None:
            self.mode = mode
        if cur_color is not None:
            self.cur_color = cur_color
        if dirty is not None:
            self.dirty = dirty
        if frame_id is not None:
            self.frame_id = frame_id
        if frame_count is not None:
            self.frame_count = frame_count

        if self.mode == MODE_MENU:
            self._str = '(d)up frame  de(l)ete frame  (s)ave  (q)uit  (0-7) color  '
        else:
            self._str = '^X for menu  ^K/^L change frame  ^V/^B change color       '
        self._str += '  frame {}/{}  {}'.format(
            self.frame_id + 1, self.frame_count,
            '*' if self.dirty else ' ')

    def get_char_and_attr(self, x, y):
        if x >= len(self._str):
            return (None, None)
        return ord(self._str[x]), 0


class ImgEditGame(game.Game):
    def __init__(self, *call_args, **call_kwargs):
        super(ImgEditGame, self).__init__(*call_args, **call_kwargs)

        self.tick_length_ms = 16  # approx. 60 tps

        self.filename = self.args.filename
        with open(self.filename, 'rb') as fh:
            self.figure = pickle.load(fh)

        self.figure.x_pos = 0
        self.figure.y_pos = 0
        self.figure.z_index = 2
        self.screen.add_figure(self.figure)

        self.border = Border(self.figure.width + 1,
                             self.figure.height + 1,
                             0, 0, 1)
        self.screen.add_figure(self.border)

        self.cursor = Cursor(z_index=3)
        self.cursor.x_pos = 0
        self.cursor.y_pos = 0
        self.cursor.visible = False
        self.screen.add_figure(self.cursor)
        self.cursor_blink_rate_ticks = 15
        self.cursor_ticks_since = 0

        self.cur_color = 0

        self.mode = MODE_EDIT
        self.userinput.set_bindings(BINDINGS_EDIT)

        self.menu_bar = MenuBar()
        self.menu_bar.x_pos = 0
        self.menu_bar.y_pos = self.figure.height + 2
        self.screen.add_figure(self.menu_bar)

        self.dirty = False

    @classmethod
    def add_args(cls, arg_parser):
        arg_parser.add_argument('filename', type=str)

    def iterate(self, ticks):
        # Animate cursor.
        if self.mode == MODE_EDIT:
            self.cursor_ticks_since += ticks
            if self.cursor_ticks_since > self.cursor_blink_rate_ticks:
                self.cursor_ticks_since = 0
                self.cursor.visible = not self.cursor.visible

        self.cursor.cur_color = curses.color_pair(self.cur_color)
        self.border.cur_color = curses.color_pair(self.cur_color)

        # Update menu bar.
        self.menu_bar.update(mode=self.mode,
                             cur_color=self.cursor.cur_color,
                             dirty=self.dirty,
                             frame_id=self.figure.cur_frame,
                             frame_count=self.figure.frame_count)

        # DEBUG: display character codes
        #s = 'LAST: {}'.format(self.userinput.last_char)
        #self.screen.stdscr.addstr(2, self.screen.width-len(s), s)

    def handle_input(self, incmd):
        if self.mode == MODE_MENU:
            if incmd == ACTION_QUIT:
                self._running = False
            elif incmd == ACTION_SAVE:
                self.save()
            elif incmd == ACTION_DUPLICATE:
                self.figure.duplicate_frame()
                self.figure.cur_frame += 1
            elif incmd == ACTION_DELETE_FRAME:
                self.figure.delete_frame()
            elif incmd == ACTION_FRAME_PREV and self.figure.cur_frame > 0:
                self.figure.cur_frame -= 1
            elif incmd == ACTION_FRAME_NEXT and self.figure.cur_frame < self.figure.frame_count - 1:
                self.figure.cur_frame += 1
            elif incmd == ACTION_COLOR_PREV and self.cur_color > 0:
                self.cur_color -= 1
            elif incmd == ACTION_COLOR_NEXT and self.cur_color < screen.MAX_COLORS - 1:
                self.cur_color += 1
            elif incmd == ACTION_COLOR_0:
                self.cur_color = 0
            elif incmd == ACTION_COLOR_1:
                self.cur_color = 1
            elif incmd == ACTION_COLOR_2:
                self.cur_color = 2
            elif incmd == ACTION_COLOR_3:
                self.cur_color = 3
            elif incmd == ACTION_COLOR_4:
                self.cur_color = 4
            elif incmd == ACTION_COLOR_5:
                self.cur_color = 5
            elif incmd == ACTION_COLOR_6:
                self.cur_color = 6
            elif incmd == ACTION_COLOR_7:
                self.cur_color = 7
            elif incmd == userinput.ESC:
                # (Exit from menu mode without running a command.)
                pass
            else:
                # No valid menu key pressed. Stay in menu mode.
                return
            # Leave menu mode.
            self.mode = MODE_EDIT
            self.userinput.set_bindings(BINDINGS_EDIT)
            return

        if self.mode == MODE_EDIT:
            if incmd == ACTION_MENU:
                self.mode = MODE_MENU
                self.userinput.set_bindings(BINDINGS_MENU)
                self.cursor.visible = False
            elif incmd == ACTION_FRAME_PREV and self.figure.cur_frame > 0:
                self.figure.cur_frame -= 1
            elif incmd == ACTION_FRAME_NEXT and self.figure.cur_frame < self.figure.frame_count - 1:
                self.figure.cur_frame += 1
            elif incmd == ACTION_COLOR_PREV and self.cur_color > 0:
                self.cur_color -= 1
            elif incmd == ACTION_COLOR_NEXT and self.cur_color < screen.MAX_COLORS - 1:
                self.cur_color += 1
            elif incmd == userinput.UP and self.cursor.y_pos > 0:
                self.cursor.y_pos -= 1
                self.cursor.visible = True
            elif incmd == userinput.DOWN and self.cursor.y_pos < self.figure.height - 1:
                self.cursor.y_pos += 1
                self.cursor.visible = True
            elif incmd == userinput.LEFT and self.cursor.x_pos > 0:
                self.cursor.x_pos -= 1
                self.cursor.visible = True
            elif incmd == userinput.RIGHT and self.cursor.x_pos < self.figure.width - 1:
                self.cursor.x_pos += 1
                self.cursor.visible = True
            elif incmd == ACTION_BACKSPACE:
                if self.cursor.x_pos > 0:
                    self.cursor.x_pos -= 1
                    self.figure.set_char_and_attr(
                        self.cursor.x_pos, self.cursor.y_pos,
                        ord(' '), self.cursor.cur_color)
                self.cursor.visible = True
                self.dirty = True
            elif incmd in TYPEABLE_CHARS:
                self.dirty = True
                self.figure.set_char_and_attr(
                    self.cursor.x_pos, self.cursor.y_pos,
                    incmd, self.cursor.cur_color)
                if self.cursor.x_pos < self.figure.width - 1:
                    self.cursor.x_pos += 1

    def save(self):
        if self.dirty:
            with open(self.filename, 'wb') as fh:
                pickle.dump(self.figure, fh)
            self.dirty = False

    def quit(self):
        self.save()


if __name__ == '__main__':
    if len(sys.argv) == 1:
        filename = raw_input('Enter a filename: ')
        if not os.path.exists(filename):
            width_str = raw_input('Enter a width: ')
            height_str = raw_input('Enter a height: ')
            frames_str = raw_input('Enter a frame count: ')
            figure = screen.DrawnFigure(width=int(width_str.strip()),
                                        height=int(height_str.strip()),
                                        frame_count=int(frames_str.strip()))
            with open(filename, 'wb') as fh:
                pickle.dump(figure, fh)
        args = [filename]
    else:
        args = sys.argv[1:]

    world = gameworld.GameWorld(ImgEditGame)
    sys.exit(world.main(args))
