#!/usr/bin/env python

import curses
import logging
import pickle
import sys

import game
import gameworld
import screen
import sound
import userinput

SLOW_WRITE_DELAY_TICKS = 2
FAST_WRITE_DELAY_TICKS = 1
FAST_WRITE_CHARS_PER_TICK = 2


# TODO: wire for sound (reference to sound pack with write(), default sound pack uses blips)
class TextDisplay(game.GameObject):
    def __init__(self, width, height, x_pos, y_pos, *args, **kwargs):
        super(TextDisplay, self).__init__(*args, **kwargs)
        self.set_figure(screen.DrawnFigure(width, height, x_pos, y_pos, 1))
        self.reset()

    def reset(self):
        self.figure.clear_frame()
        self.cursor_x = 0
        self.cursor_y = 0
        self.string_buffer = []
        self.string_pos = 0
        self.ticks_since_last = 0
        self.make_this_faster = False

    def _write_char(self, char, attr):
        if self.cursor_y >= self.figure.height:
            return

        if char == 13:
            self.cursor_x = 0
            self.cursor_y += 1
        else:
            self.figure.set_char_and_attr(
                self.cursor_x, self.cursor_y, char, attr)
            self.cursor_x += 1
        if self.cursor_x >= self.figure.width:
            self.cursor_y += 1
            self.cursor_x = 0

    def is_done(self):
        return not self.string_buffer

    def iterate(self, ticks):
        if not self.string_buffer:
            return
        if self.string_pos >= len(self.string_buffer[0][0]):
            self.string_buffer.pop(0)
            self.string_pos = 0
            return

        self.ticks_since_last += ticks
        if (self.ticks_since_last > (FAST_WRITE_DELAY_TICKS
                                     if self.make_this_faster
                                     else self.string_buffer[0][2])):
            self.ticks_since_last = 0
            non_whitespace_to_write = (FAST_WRITE_CHARS_PER_TICK
                                       if self.make_this_faster
                                       else 1)
            while (non_whitespace_to_write > 0 and
                   self.string_pos < len(self.string_buffer[0][0])):
                c = self.string_buffer[0][0][self.string_pos]
                if c == '\n':
                    c = 13
                elif c == ' ':
                    c = 32
                else:
                    c = ord(c)
                    non_whitespace_to_write -= 1
                self._write_char(c,
                                 self.string_buffer[0][1])
                self.string_pos += 1
            # TODO: sound

    def write(self, s, attr=0,
              char_delay_ticks=SLOW_WRITE_DELAY_TICKS):
        self.string_buffer.append([s, attr, char_delay_ticks])


class StartScreenDisplay(game.GameObject):
    def __init__(self, *args, **kwargs):
        super(StartScreenDisplay, self).__init__(*args, **kwargs)
        self.set_figure(screen.load_figure('lostcat/start.dat'))

        self.animation_delay_ticks = 10
        self.ticks_since_last = 0

    def iterate(self, ticks):
        self.ticks_since_last += ticks
        if self.ticks_since_last > self.animation_delay_ticks:
            self.ticks_since_last = 0
            self.figure.cur_frame = (self.figure.cur_frame + 1) % self.figure.frame_count


MAIN_MENU_OPTIONS = [
    ('Continue', 'closemenu'),
    ('Back to title', 'restart'),
    ('Quit', 'quit'),
]

class MenuDisplay(game.GameObject):
    def __init__(self, *args, **kwargs):
        super(MenuDisplay, self).__init__(*args, **kwargs)
        self.options = kwargs.get('options', MAIN_MENU_OPTIONS)
        self.border_char = kwargs.get('border_char', ord('#'))
        self.border_attr = kwargs.get('border_attr', curses.color_pair(1))

        longest_len = 0
        for x in self.options:
            if len(x[0]) > longest_len:
                longest_len = len(x[0])

        width = longest_len + 8
        height = len(self.options) * 2 + 3
        x_pos = 40 - (width/2)
        y_pos = 12 - (height/2)
        self.set_figure(screen.DrawnFigure(width, height, x_pos, y_pos, z_index=1))

        # Draw border.
        for i in xrange(width):
            self.figure.set_char_and_attr(i, 0, self.border_char, self.border_attr)
            self.figure.set_char_and_attr(i, height-1, self.border_char, self.border_attr)
        for j in xrange(height):
            self.figure.set_char_and_attr(0, j, self.border_char, self.border_attr)
            self.figure.set_char_and_attr(width-1, j, self.border_char, self.border_attr)

        # Draw option text.
        for i, o in enumerate(self.options):
            self.figure.write_str(4, (i+1) * 2, o[0], 0)
            
        self.selected = 0

        self.indicator_visible = True
        self.animation_delay_ticks = 10
        self.ticks_since_last = 0

    def iterate(self, ticks):
        self.ticks_since_last += ticks
        if self.ticks_since_last > self.animation_delay_ticks:
            self.ticks_since_last = 0
            self.indicator_visible = not self.indicator_visible
            self.set_indicator(on=self.indicator_visible)

    def set_selected(self, selected):
        if selected < 0 or selected > len(self.options) - 1:
            return
        self.set_indicator(on=False)
        self.selected = selected

    def get_selected(self):
        return self.options[self.selected]

    def set_indicator(self, on=True):
        y = (self.selected + 1) * 2
        self.figure.set_char_and_attr(
            2, y, ord('>') if on else ord(' '), 0)
        self.figure.set_char_and_attr(
            5 + len(self.options[self.selected][0]), y,
            ord('<') if on else ord(' '), 0)


class ActiveObject(game.GameObject):
    def __init__(self, *args, **kwargs):
        super(ActiveObject, self).__init__(*args, **kwargs)
        self.x = kwargs.get('x')
        self.y = kwargs.get('y')
        self.figure = None

    def collides_with(self, other):
        return ((self.figure.x_pos < other.figure.x_pos + other.figure.width) and
                (self.figure.y_pos < other.figure.y_pos + other.figure.height) and
                (other.figure.x_pos < self.figure.x_pos + self.figure.width) and
                (other.figure.y_pos < self.figure.y_pos + self.figure.height))

    def handle_player_collision(self, game):
        pass

class Player(ActiveObject):
    def __init__(self, *args, **kwargs):
        super(Player, self).__init__(*args, **kwargs)
        self.figure = screen.load_figure('lostcat/player.dat')
        self.figure.x_pos = self.x
        self.figure.y_pos = self.y

class Shovel(ActiveObject):
    def __init__(self, *args, **kwargs):
        super(Shovel, self).__init__(*args, **kwargs)
        self.figure = screen.load_figure('lostcat/shovel.dat')
        self.figure.x_pos = self.x
        self.figure.y_pos = self.y

    def handle_player_collision(self, game):
        # TODO: shovel cutscene
        logging.info('Player picked up shovel')
        self.set_active(active=False)
        game.digger_display.remove_active_object('shovel')
        game.player_power = 1

class Barn(ActiveObject):
    def __init__(self, *args, **kwargs):
        super(Barn, self).__init__(*args, **kwargs)
        self.figure = screen.load_figure('lostcat/barn.dat')
        self.figure.x_pos = self.x
        self.figure.y_pos = self.y

    def handle_player_collision(self, game):
        # TODO: barn sequence
        game.set_mode(MODE_BARN)
        logging.info('Player touched barn')
        pass

class MyCat(ActiveObject):
    def __init__(self, *args, **kwargs):
        super(MyCat, self).__init__(*args, **kwargs)
        self.figure = screen.load_figure('lostcat/mycat.dat')
        self.figure.x_pos = self.x
        self.figure.y_pos = self.y

    def handle_player_collision(self, game):
        # TODO: my cat cutscene; my cat removed; (end of game, until we do more)
        logging.info('Player touched my cat')
        pass


LEVEL_EMPTY = 0
LEVEL_DIRT_1 = 101
LEVEL_DIRT_2 = 102
LEVEL_DIRT_3 = 103
LEVEL_DIRT_4 = 104
LEVEL_DIRT_5 = 105
LEVEL_COIN_1 = 201
LEVEL_COIN_2 = 202
LEVEL_COIN_3 = 203
LEVEL_COIN_4 = 204
LEVEL_COIN_5 = 205

# (char, color number, other attrs)
DIRT_CHARS = {
    LEVEL_EMPTY: (ord(' '), 0, 0),
    LEVEL_DIRT_1: (ord('.'), 2, 0),
    LEVEL_DIRT_2: (ord('.'), 1, 0),
    LEVEL_DIRT_3: (ord('~'), 1, 0),
    LEVEL_DIRT_4: (ord('~'), 0, 0),
    LEVEL_DIRT_5: (ord('#'), 6, 0),
    LEVEL_COIN_1: (ord('1'), 3, curses.A_BOLD),
    LEVEL_COIN_2: (ord('2'), 3, curses.A_BOLD),
    LEVEL_COIN_3: (ord('3'), 3, curses.A_BOLD),
    LEVEL_COIN_4: (ord('4'), 3, curses.A_BOLD),
    LEVEL_COIN_5: (ord('5'), 3, curses.A_BOLD),
}


class LevelFigure(screen.Figure):
    def __init__(self, *args, **kwargs):
        super(LevelFigure, self).__init__(*args, **kwargs)

        self.active_objects = {}
        self.dirt = []
        for j in range(self.height):
            self.dirt.append([0] * self.width)

    def add_active_object(self, name, classname, x, y):
        self.active_objects[name] = (classname, x, y)

    def set_dirt(self, val, x, y):
        assert y < len(self.dirt)
        assert x < len(self.dirt[0])
        self.dirt[y][x] = val

    def get_dirt(self, x, y):
        return self.dirt[y][x]

    def get_char_and_attr(self, x, y):
        (char, color, other_attrs) = DIRT_CHARS[self.dirt[y][x]]
        return char, curses.color_pair(color) | other_attrs


SCROLL_TOP = 5
SCROLL_BOTTOM = 15

class DiggerDisplay(game.GameObject):
    def __init__(self, *args, **kwargs):
        super(DiggerDisplay, self).__init__(*args, **kwargs)
        self.figure = kwargs.get('level_data')
        self.figure.x_pos = 0
        self.figure.y_pos = 1
        self.figure.z_index = 1

        self.active_objects = {}
        self.active_objects_to_remove = []
        mod = sys.modules[__name__]
        for name in self.figure.active_objects:
            (classname, x, y) = self.figure.active_objects[name]
            ctr = getattr(mod, classname)
            self.active_objects[name] = ctr(x=x + self.figure.x_pos, y=y + self.figure.y_pos)
            self.active_objects[name].figure.z_index = self.figure.z_index
            self.game.add_object(self.active_objects[name])

        self.animation_delay_ticks = 10
        self.ticks_since_last = 0

    def iterate(self, ticks):
        self.ticks_since_last += ticks
        if self.ticks_since_last > self.animation_delay_ticks:
            self.ticks_since_last = 0
            for name in self.active_objects:
                f = self.active_objects[name].figure
                f.cur_frame = (f.cur_frame + 1) % f.frame_count

        for name in self.active_objects_to_remove:
            self.active_objects[name].set_active(active=False)
            del self.active_objects[name]
        self.active_objects_to_remove = []

    def set_active(self, active=True):
        super(DiggerDisplay, self).set_active(active=active)
        for name in self.active_objects:
            self.active_objects[name].set_active(active=active)

    def move_player(self, x_offset=0, y_offset=0):
        assert self.game is not None
        assert 'player' in self.active_objects

        p = self.active_objects['player']

        # Disallow moving off the level (no indication).
        level_pos_x = p.figure.x_pos - self.figure.x_pos + x_offset
        level_pos_y = p.figure.y_pos - self.figure.y_pos + y_offset
        if (level_pos_x < 0 or level_pos_x + p.figure.width > self.figure.width or
            level_pos_y < 0 or level_pos_y + p.figure.height > self.figure.height):
            return

        edge_dirt = []
        dirt_coords = []
        for i in xrange(p.figure.width):
            for j in xrange(p.figure.height):
                x = level_pos_x + i
                y = level_pos_y + j
                dirt = self.figure.get_dirt(x, y)
                if dirt > 0:
                    edge_dirt.append(dirt)
                    dirt_coords.append((x, y))

        # Test for dirt collisions.  Disallow if not enough power, indicate disallowed.
        enough_power = True
        coin_total = 0
        for dirt in edge_dirt:
            if (LEVEL_DIRT_1 <= dirt <= LEVEL_DIRT_1 + 99) and (dirt > self.game.player_power + 100):
                enough_power = False
                break
            if LEVEL_COIN_1 <= dirt <= LEVEL_COIN_1 + 99:
                coin_total += dirt - LEVEL_COIN_1 + 1

        if not enough_power:
            if not self.game.seen_cant_dig:
                # TODO: "need shovel" cutscene
                self.game.seen_cant_dig = True
            # TODO: Effect: Digging disallowed.
            logging.info('Player not powerful enough to move that way')
            pass

        else:
            if coin_total > 0:
                # TODO: indicate coin collection (sound)
                logging.info('Player collected %r coins', coin_total)
                self.game.coins_collected += coin_total

            elif len(dirt_coords) > 0:
                # TODO: indicate digging (sound).
                logging.info('Player dug dirt')

            else:
                # Moved through air.  (No indicator planned.)
                pass

            # Move player.
            p.figure.x_pos += x_offset
            p.figure.y_pos += y_offset

            # Clear dirt.
            for x, y in dirt_coords:
                self.figure.set_dirt(LEVEL_EMPTY, x, y)

            # Test for action object collisions.  Execute routines.
            for name in self.active_objects:
                if name == 'player':
                    continue
                if p.collides_with(self.active_objects[name]):
                    self.active_objects[name].handle_player_collision(self.game)

            # Scroll digger display if necessary.
            if p.figure.y_pos < SCROLL_TOP:
                self.vert_scroll(-1)
            elif p.figure.y_pos + p.figure.height > SCROLL_BOTTOM:
                self.vert_scroll(1)

    def vert_scroll(self, y_offset):
        new_level_y = self.figure.y_pos - y_offset
        
        # Don't scroll objects to the point where the level is off the screen.
        if new_level_y > 1 or new_level_y + self.figure.height < 23:
            return

        self.figure.y_pos = new_level_y
        for name in self.active_objects:
            self.active_objects[name].figure.y_pos -= y_offset

    def remove_active_object(self, name):
        if name not in self.active_objects:
            return
        self.active_objects_to_remove.append(name)


class CoinFigure(screen.Figure):
    def __init__(self, *args, **kwargs):
        super(CoinFigure, self).__init__(*args, **kwargs)
        self.str = ' ' * 80

    def get_char_and_attr(self, x, y):
        return ord(self.str[x]), 0

    def update(self, coins):
        s = '{} coins collected'.format(coins)
        self.str = (' ' * (80-len(s))) + s


class CoinDisplay(game.GameObject):
    def __init__(self, *args, **kwargs):
        super(CoinDisplay, self).__init__(*args, **kwargs)
        self.figure = CoinFigure(80, 1, 0, 0, z_index=10)
        self.last_coins = 0

    def iterate(self, ticks):
        if self.game.coins_collected != self.last_coins:
            self.figure.update(self.game.coins_collected)
            self.last_coins = self.game.coins_collected


MAX_POWER = 5
POWER_COST = {
    1: 10,
    2: 20,
    3: 40,
    4: 60,
    5: 80,
}

class BarnDisplay(game.GameObject):
    def __init__(self, *args, **kwargs):
        super(BarnDisplay, self).__init__(*args, **kwargs)
        # HACK: load the barn from disk, and draw text on top of it at fixed coords
        self.set_figure(screen.load_figure('lostcat/barnscreen.dat'))
        self.figure.z_index = 1
        self.text_display = TextDisplay(40, 10, 10, 10)
        self.text_display.figure.z_index = 2
        self.game.add_object(self.text_display)
        self.reset()

        self.prompted = False

    def set_active(self, active=True):
        super(BarnDisplay, self).set_active(active=active)
        self.text_display.set_active(active=active)

    def reset(self):
        self.text_display.reset()

        next_cost = POWER_COST.get(self.game.player_power + 1, 99999)
        if self.game.player_power == 0:
            self.text_display.write('Hey there kid.\n\nI sell shovels here, but only by\ntrade in.\n\nGo find a cheap shovel and get some\nmoney.  Good luck!')
            self.prompted = False
        elif self.game.player_power == MAX_POWER:
            # No more upgrades.
            self.text_display.write('You have the best tool I got, kid!\n\nGo get that cat!')
            self.prompted = False
        elif self.game.coins_collected < next_cost:
            # Can't afford.
            self.text_display.write('Hey there!\n\nI can upgrade your shovel, but it\'ll\n'
                                    'cost you ' + str(next_cost) + ' coins.\n\n'
                                    'Go find some more money.')
            self.prompted = False
        else:
            # Can upgrade.
            self.text_display.write('Hiyo!\n\nWould you like to buy a better shovel?  The price is ' +
                                    str(next_cost) + ' coins.\n\n[Y]es or [N]o?')
            self.prompted = True

    def yes(self):
        if not self.text_display.is_done():
            return
        if not self.prompted:
            return
        self.game.player_power += 1
        self.game.coins_collected -= POWER_COST[self.game.player_power]
        self.game.set_mode(MODE_GAME)

    def no(self):
        if not self.text_display.is_done():
            return
        if not self.prompted:
            return
        self.game.set_mode(MODE_GAME)

    def select(self):
        if not self.text_display.is_done():
            self.text_display.make_this_faster = True
            return
        if self.prompted:
            return
        self.game.set_mode(MODE_GAME)


MODE_INIT = 1
MODE_START = 2
MODE_CUTSCENE = 3
MODE_MENU = 4
MODE_GAME = 5
MODE_BARN = 6

BINDINGS = {
    MODE_INIT: {
        32: 'ack_init',
        10: 'ack_init',
        curses.KEY_ENTER: 'ack_init',
    },
    MODE_START: {
        27: 'menu',
        32: 'start',
        10: 'start',
        curses.KEY_ENTER: 'start',
    },
    MODE_MENU: {
        27: 'closemenu',
        32: 'select',
        10: 'select',
        curses.KEY_ENTER: 'select',
        curses.KEY_UP: 'prev',
        curses.KEY_DOWN: 'next',
    },
    MODE_GAME: {
        27: 'menu',
        curses.KEY_UP: 'up',
        curses.KEY_DOWN: 'down',
        curses.KEY_LEFT: 'left',
        curses.KEY_RIGHT: 'right',
        ord('w'): 'up',
        ord('s'): 'down',
        ord('a'): 'left',
        ord('d'): 'right',
    },
    MODE_CUTSCENE: {
        27: 'menu',
        32: 'action',
        10: 'action',
        curses.KEY_ENTER: 'action',
    },
    MODE_BARN: {
        27: 'menu',
        32: 'select',
        10: 'select',
        curses.KEY_ENTER: 'select',
        ord('y'): 'yes',
        ord('n'): 'no',
    }
}


class LostCatGame(game.Game):
    SCREEN_WIDTH = 80
    SCREEN_HEIGHT = 25

    def __init__(self, *args, **kwargs):
        super(LostCatGame, self).__init__(*args, **kwargs)

        # TODO: move game data to a reset function
        self.player_power = 0
        self.coins_collected = 0
        self.seen_cant_dig = False

        self.tick_length_ms = 16
        self.game_mode = None

        self.last_cmd = 'none'

        self.cur_menu_obj = None
        self.menu_prev_mode = None

        # TODO: init sound
        self.sound = sound.Sound()

        self.init_screen = TextDisplay(79, 20, 0, 0)
        self.init_screen.make_this_faster = True
        self.init_screen.write(
            '\n                                 L O S T   C A T\n\n\n\n')
        if self.sound.is_enabled():
            self.init_screen.write(
                ' Sound is available.  Click on this window, then press SPACE.')
        else:
            #self.init_screen.write(
            #    ' Sound is not available.  To enable sound, please install\n'
            #    ' PyGame (www.pygame.org).\n\n Press SPACE to continue.')
            self.init_screen.write(
                ' Lost Cat v0.1, Ludum Dare submission edition.\n\n Some stuff is missing.  Stay tuned for v0.2.  :)\n\n Press SPACE to continue.\n During the game, press ESC for the menu.')
        self.add_object(self.init_screen)

        self.start_screen = StartScreenDisplay()
        self.add_object(self.start_screen)

        self.main_menu_display = MenuDisplay(options=MAIN_MENU_OPTIONS)
        self.add_object(self.main_menu_display)

        with open('lostcat/level.dat', 'rb') as fh:
            level_data = pickle.load(fh)
        self.digger_display = DiggerDisplay(game=self, level_data=level_data)
        self.digger_display.set_active(active=False)
        self.add_object(self.digger_display)

        self.coin_display = CoinDisplay(game=self)
        self.add_object(self.coin_display)

        self.barn_display = BarnDisplay(game=self)
        self.add_object(self.barn_display)

        self.set_mode(MODE_INIT)

    def iterate(self, ticks):
        super(LostCatGame, self).iterate(ticks)

        # TODO: DEBUG
        #s = 'mode:{} last:{}'.format(self.game_mode, self.last_cmd)
        #self.screen.stdscr.addstr(1, self.screen.width-len(s), s)

    def handle_input(self, incmd):
        if incmd:
            self.last_cmd = incmd

        if incmd == 'quit':
            self.signal_quit()

        elif incmd == 'ack_init':
            self.set_mode(MODE_START)

        elif incmd == 'start':
            self.set_mode(MODE_GAME)

        elif incmd == 'restart':
            # TODO: reset game state!
            self.set_mode(MODE_START)

        elif incmd == 'menu':
            self.cur_menu_obj = self.main_menu_display
            self.menu_prev_mode = self.game_mode
            self.set_mode(MODE_MENU)
        elif incmd == 'closemenu':
            self.set_mode(self.menu_prev_mode)
        elif self.game_mode == MODE_MENU and incmd == 'select':
            self.handle_input(self.cur_menu_obj.get_selected()[1])
        elif incmd == 'prev':
            self.cur_menu_obj.set_selected(self.cur_menu_obj.selected - 1)
        elif incmd == 'next':
            self.cur_menu_obj.set_selected(self.cur_menu_obj.selected + 1)

        elif incmd == 'up':
            self.digger_display.move_player(y_offset=-1)
        elif incmd == 'down':
            self.digger_display.move_player(y_offset=1)
        elif incmd == 'left':
            self.digger_display.move_player(x_offset=-1)
        elif incmd == 'right':
            self.digger_display.move_player(x_offset=1)

        elif self.game_mode == MODE_BARN:
            if incmd == 'yes':
                self.barn_display.yes()
            elif incmd == 'no':
                self.barn_display.no()
            elif incmd == 'select':
                self.barn_display.select()

    def set_mode(self, mode):
        prev_mode = self.game_mode
        self.game_mode = mode

        # Deactivate previous mode.
        if prev_mode == MODE_INIT:
            self.init_screen.set_active(active=False)

        elif prev_mode == MODE_START:
            self.start_screen.set_active(active=False)
            # TODO: stop start-screen music

        elif prev_mode == MODE_MENU:
            self.main_menu_display.set_active(active=False)

        elif prev_mode == MODE_GAME:
            self.digger_display.set_active(active=False)
            self.coin_display.set_active(active=False)

        elif prev_mode == MODE_BARN:
            p = self.digger_display.active_objects['player']
            p.figure.x_pos = 68
            p.figure.y_pos = 4
            self.barn_display.set_active(active=False)

        # Set new bindings.
        self.userinput.set_bindings(BINDINGS[mode])

        # Activate new mode.
        if self.game_mode == MODE_INIT:
            self.init_screen.set_active(active=True)

        elif self.game_mode == MODE_START:
            for o in self.game_objects:
                o.set_active(active=False)
            self.start_screen.set_active(active=True)
            # TODO: start start-screen music

        elif self.game_mode == MODE_MENU:
            self.main_menu_display.set_selected(0)
            self.main_menu_display.set_active(active=True)

        elif self.game_mode == MODE_GAME:
            self.digger_display.set_active(active=True)
            self.coin_display.set_active(active=True)

        elif self.game_mode == MODE_BARN:
            self.barn_display.reset()
            self.barn_display.set_active(active=True)


if __name__ == '__main__':
    world = gameworld.GameWorld(LostCatGame)
    sys.exit(world.main(sys.argv[1:]))
